@@ -11,6 +11,10 @@ import requests
 from devo.common import default_from, default_to
 from .processors import processors, proc_json, \
     json_compact_simple_names, proc_json_compact_simple_to_jobj
+import calendar
+from datetime import datetime, timedelta
+import pytz
+
 
 CLIENT_DEFAULT_APP_NAME = 'python-sdk-app'
 CLIENT_DEFAULT_USER = 'python-sdk-user'
@@ -29,12 +33,19 @@ SIMPLECOMPACT_TO_ARRAY = "jsoncompactsimple_to_array"
 
 ERROR_MSGS = {
     "no_query": "Error: Not query provided.",
-    "no_auth": "Client dont have key&secret or auth token/jwt",
+    "no_auth": "Client don't have key&secret or auth token/jwt",
     "no_endpoint": "Endpoint 'address' not found",
     "to_but_no_from": "If you use end dates for the query 'to' it is "
                       "necessary to use start date 'from'",
-    "binary_format_requires_output": "Binary format like `msgpack` and `xls` "
-                                     "requires output parameter"
+    "binary_format_requires_output": "Binary format like `msgpack` and `xls` requires output parameter",
+    "wrong_processor": "processor must be lambda/function or one of the defaults API processors.",
+    "default_keepalive_only": "Mode '%s' always uses default KeepAlive Token",
+    "keepalive_not_supported": "Mode '%s' does not support KeepAlive Token",
+    "stream_mode_not_supported": "Mode '%s' does not support stream mode",
+    "future_queries_not_supported": "Modes 'xls' and 'msgpack' does not support future queries because KeepAlive"
+                                    " tokens are not available for those resonses type",
+    "missing_api_key": "You need a API Key and API secret to make this",
+    "data_query_error": "Error while receiving query data: %s "
 }
 
 DEFAULT_KEEPALIVE_TOKEN = '\n'
@@ -45,27 +56,70 @@ NO_KEEPALIVE_TOKEN = None
 class DevoClientException(Exception):
     """ Default Devo Client Exception """
 
+    def __init__(self, message, status=None, code=None, cause=None):
+        if isinstance(message, dict):
+            self.status = message.get('status', status)
+            self.cause = message.get('cause', cause)
+            self.message = message.get('msg',
+                                       message if isinstance(message, str)
+                                       else json.dumps(message))
+            self.cid = message.get('cid', None)
+            self.code = message.get('code', code)
+            self.timestamp = message.get('timestamp',
+                                         time.time_ns() // 1000000)
+        else:
+            self.message = message
+            self.status = status
+            self.cause = cause
+            self.cid = None
+            self.code = code
+            self.timestamp = time.time_ns() // 1000000
+        super().__init__(message)
 
-def raise_exception(error):
-    """Util function for raise exceptions"""
-    try:
-        if isinstance(error, str):
-            raise DevoClientException(proc_json()(error))
-        if isinstance(error, DevoClientException):
-            if isinstance(error.args[0], str):
-                raise DevoClientException(proc_json()(error.args[0]))
-            raise DevoClientException(error.args[0])
-        if isinstance(error, dict):
-            raise DevoClientException(error)
-        response_text = proc_json()(error.text)
-        raise DevoClientException(response_text)
-    except json.decoder.JSONDecodeError:
-        raise DevoClientException(_format_error(error))
+    def __str__(self):
+        return self.message + ((": " + self.cause) if self.cause else '')
 
 
-def _format_error(error):
-    return '{"msg": "Error Launching Query", "status": 500, ' \
-           '"object": "%s"}' % str(error).replace("\"", "\\\"")
+def raise_exception(error_data, status=None):
+    if isinstance(error_data, requests.models.Response):
+        raise DevoClientException(
+            _format_error(error_data.json(), status=error_data.status_code))
+
+    elif isinstance(error_data, str):
+        if not status:
+            raise DevoClientException(
+                _format_error({"object": error_data}, status=None))
+        raise DevoClientException(
+            _format_error({"object": error_data}, status=status))
+    elif isinstance(error_data, BaseException):
+        raise DevoClientException(_format_error(error_data, status=None))\
+            from error_data
+    else:
+        raise DevoClientException(_format_error(error_data, status=None))
+
+
+def _format_error(error, status):
+    if isinstance(error, dict):
+        response = {
+            "msg": error.get("msg", "Error Launching Query"),
+            "cause": error.get("object") or error.get("context") or error
+        }
+        # 'object' may be a list
+        if isinstance(response["cause"], list):
+            response["cause"] = ": ".join(response["cause"])
+        if status:
+            response['status'] = status
+        elif 'status' in error:
+            response['status'] = error['status']
+        for item in ['code', 'cid', 'timestamp']:
+            if item in error:
+                response[item] = error[item]
+        return response
+    else:
+        return {
+            "msg": str(error),
+            "cause": str(error)
+        }
 
 
 class ClientConfig:
@@ -120,8 +174,7 @@ class ClientConfig:
             self.proc = "CUSTOM"
             self.processor = processor
         else:
-            raise_exception("processor must be lambda/function or one of"
-                            "the defaults API processors.")
+            raise_exception(ERROR_MSGS["wrong_processor"])
         return True
 
     def set_user(self, user=CLIENT_DEFAULT_USER):
@@ -148,29 +201,28 @@ class ClientConfig:
         :param keepAliveToken: KeepAlive token for long responses queries
         :return:
         """
-        if self.response in ['xls']:
-            self.keepAliveToken = DEFAULT_KEEPALIVE_TOKEN
-            if keepAliveToken not in [DEFAULT_KEEPALIVE_TOKEN]:
-                logging.warning(f"Mode '{self.response}' only supports default"
-                                f" KeepAlive Token")
-        elif self.response in ['msgpack']:
+        # The KeepAlive token does not apply to any format other than 'csv',
+        # 'tsv'. All json* responses always use default b'    ' token for
+        # keepalive (cannot be modified), but implementation uses
+        # NO_KEEP_ALIVE value as it does not change the query msgpack and
+        # xls does not support keepalive
+        if self.response in ['json', 'json/compact', 'json/simple',
+                             'json/simple/compact']:
             self.keepAliveToken = NO_KEEPALIVE_TOKEN
             if keepAliveToken not in [NO_KEEPALIVE_TOKEN,
                                       DEFAULT_KEEPALIVE_TOKEN]:
-                logging.warning(f"Mode '{self.response}' does not support"
-                                f" KeepAlive Token")
-        elif self.response not in ['csv', 'tsv']:
-            self.keepAliveToken = NO_KEEPALIVE_TOKEN
-            if keepAliveToken not in [NO_KEEPALIVE_TOKEN,
-                                      DEFAULT_KEEPALIVE_TOKEN]:
-                logging.warning(f"Mode '{self.response}' does not support"
-                                f" custom KeepAlive Token")
-        elif keepAliveToken == NO_KEEPALIVE_TOKEN:
-            self.keepAliveToken = DEFAULT_KEEPALIVE_TOKEN
-            logging.warning(f"Mode '{self.response}' requires KeepAlive Token,"
-                            f" default token configured")
-        else:
+                logging.warning(
+                    ERROR_MSGS["default_keepalive_only"] % self.response)
+        # In the cases 'csv', 'tsv' you can use any value passed in
+        # 'keepAliveToken'.
+        elif self.response in ['csv', 'tsv']:
             self.keepAliveToken = keepAliveToken
+        else:
+            if keepAliveToken not in [NO_KEEPALIVE_TOKEN,
+                                      DEFAULT_KEEPALIVE_TOKEN]:
+                logging.warning(
+                    ERROR_MSGS["keepalive_not_supported"] % self.response)
+            self.keepAliveToken = NO_KEEPALIVE_TOKEN
         return True
 
 
@@ -180,13 +232,15 @@ class Client:
     """
 
     def __init__(self, address=None, auth=None, config=None,
-                 retries=None, timeout=None, verify=None):
+                 retries=None, timeout=None, verify=None, retry_delay=None):
         """
         Initialize the API with this params, all optionals
         :param address: endpoint
         :param auth: object with auth params (key, secret, token, jwt)
         :param retries: number of retries for a query
-        :param timeout: timeout of socket
+        :param timeout: timeout of socket. Default: None (blocking queries)
+        :param retry_delay: delay to wait between retry attemps, exponential
+         backoff algorithm applied with rate reduction of 2. Default: 5 seconds
         :param verify: Whether enable or disable the TLS authentication of
          endpoint
         """
@@ -208,19 +262,20 @@ class Client:
             retries = retries if retries is not None \
                 else config.get("retries", 0)
             timeout = timeout if timeout is not None \
-                else config.get("timeout", 30)
+                else config.get("timeout", 300)
+            retry_delay = retry_delay if retry_delay is not None \
+                else config.get("retry_delay", 5)
             self.config = self._from_dict(config)
 
         self.auth = auth
         if not address:
-            raise raise_exception(
-                _format_error(ERROR_MSGS['no_endpoint'])
-            )
+            raise raise_exception(ERROR_MSGS['no_endpoint'])
 
         self.address = self.__get_address_parts(address)
 
         self.retries = int(retries) if retries else 0
-        self.timeout = int(timeout) if timeout else 30
+        self.timeout = int(timeout) if timeout else 300
+        self.retry_delay = int(retry_delay) if retry_delay else 5
         self.verify = verify if verify is not None else True
 
         # For internal testing purposes, Devo will never expose a REST service
@@ -352,8 +407,19 @@ class Client:
             if not dates['to']:
                 dates['to'] = "now()"
             if self.config.stream:
-                logging.warning(f"Mode '{self.config.response}' does not"
-                                f" support stream mode")
+                logging.warning(ERROR_MSGS["stream_mode_not_supported"] % self.config.response)
+            # If is a future query and response type is 'xls' or 'msgpack'
+            # return warning because is not available.
+            if self._future_queries_available(self.config.response):
+                self.config.stream = False
+            else:
+
+                fromDate = self._fromDate_parser(default_to(dates['from']))
+                toDate = self._toDate_parser(fromDate, default_to(dates['to']))
+
+                if toDate > default_to("now()"):
+                    raise raise_exception(ERROR_MSGS["future_queries_not_supported"])
+
             self.config.stream = False
 
         return self._call(
@@ -367,27 +433,39 @@ class Client:
         :return: Response from API
         """
         if self.config.stream:
-            return self._return_stream(payload)
-        response = self._make_request(payload)
+            return self._return_string_stream(payload)
+        # We access to the whole server response value
+        (wholeResponse, _, _) = self._make_request(payload)
 
-        if isinstance(response, str):
-            return proc_json()(response)
+        if isinstance(wholeResponse, str):
+            # Analyse the response data to check if there are any error
+            # messages within the response.
+            self._error_handler(wholeResponse)
+            return proc_json()(wholeResponse)
         if self.config.response in ["msgpack", "xls"]:
-            return self.config.processor(response.content)
+            # Analyse the response data to check if there are any error
+            # messages within the response.
+            self._error_handler(wholeResponse.content)
+            return self.config.processor(wholeResponse.content)
         else:
+            # Analyse the response data to check if there are any error
+            # messages within the response.
+            self._error_handler(wholeResponse.text)
             return self.config.processor(
-                self._keepalive_content_sanitize(response.text))
+                self._keepalive_content_sanitize(wholeResponse.text))
 
-    def _return_stream(self, payload):
+    def _return_string_stream(self, payload):
         """If it's a stream call, return yield lines
         :param payload: item with headers for request
         :return line: yield-generator item
         """
-        raw_response = self._make_request(payload)
+        # We access to the iterLines response from the server.
+        (_, iterStringResponse, _) = self._make_request(payload)
 
         response = filter(lambda l: self._empty_lines_sanitize(l),
                           map(lambda l: self._keepalive_stream_sanitize(l),
-                              map(lambda l: l.decode('utf-8'), raw_response)))
+                              map(lambda l: l.decode('utf-8'),
+                                  iterStringResponse)))
         try:
             first = next(response)
         except StopIteration:
@@ -404,15 +482,20 @@ class Client:
             else:
                 yield self.config.processor(first)
             for line in response:
+                # Analyse the response data to check if there are any error
+                # messages within the response.
+                self._error_handler(line)
+
                 yield self.config.processor(line)
         else:
             yield proc_json()(first)
 
-    def _empty_lines_sanitize(self, line):
+    @staticmethod
+    def _empty_lines_sanitize(line):
         return line.strip()
 
     def _keepalive_content_sanitize(self, response):
-        if self.config.keepAliveToken == NO_KEEPALIVE_TOKEN or\
+        if self.config.keepAliveToken == NO_KEEPALIVE_TOKEN or \
                 self.config.keepAliveToken is None:
             return response
         elif self.config.keepAliveToken == DEFAULT_KEEPALIVE_TOKEN:
@@ -434,7 +517,7 @@ class Client:
             return response.replace(f'{self.config.keepAliveToken}', '')
 
     def _keepalive_stream_sanitize(self, line):
-        if self.config.keepAliveToken == NO_KEEPALIVE_TOKEN or\
+        if self.config.keepAliveToken == NO_KEEPALIVE_TOKEN or \
                 self.config.keepAliveToken is None:
             return line
         elif self.config.keepAliveToken == DEFAULT_KEEPALIVE_TOKEN:
@@ -463,32 +546,42 @@ class Client:
         tries = 0
         while tries <= self.retries:
             try:
-                response = requests.post("{}://{}".format(
-                    "http" if self.unsecure_http else "https",
-                    "/".join(self.address)),
-                    data=payload,
-                    headers=self._get_headers(payload),
-                    verify=self.verify,
-                    timeout=self.timeout,
-                    stream=self.config.stream)
+                response = self.__request(payload)
                 if response.status_code != 200:
                     raise DevoClientException(response)
 
                 if self.config.stream:
-                    return response.iter_lines()
-                return response
+                    if (self.config.response in ["msgpack", "xls"]):
+                        return (None, None, response.iter_content())
+                    return (None, response.iter_lines(), None)
+                # In case of NOT stream mode we return the whole server
+                # response.
+                return (response, None, None)
             except requests.exceptions.ConnectionError as error:
                 tries += 1
                 if tries > self.retries:
-                    return raise_exception(_format_error(error))
-                time.sleep(self.timeout)
+                    return raise_exception(error)
+                time.sleep(self.retry_delay * (2 ** (tries-1)))
             except DevoClientException as error:
                 if isinstance(error, DevoClientException):
                     raise_exception(error.args[0])
                 else:
                     raise_exception(error)
             except Exception as error:
-                return raise_exception(_format_error(error))
+                return raise_exception(error)
+
+    def __request(self, payload):
+        """
+        POST request method extracted for mocking purposes
+        """
+        return requests.post("{}://{}".format(
+            "http" if self.unsecure_http else "https",
+            "/".join(self.address)),
+            data=payload,
+            headers=self._get_headers(payload),
+            verify=self.verify,
+            timeout=self.timeout,
+            stream=self.config.stream)
 
     @staticmethod
     def _get_payload(query, query_id, dates, opts):
@@ -533,7 +626,7 @@ class Client:
         if opts["destination"]:
             payload['destination'] = opts['destination']
 
-        if opts["keepAliveToken"] is not None and\
+        if opts["keepAliveToken"] is not None and \
                 opts["keepAliveToken"] != NO_KEEPALIVE_TOKEN:
             if opts["keepAliveToken"] == EMPTY_EVENT_KEEPALIVE_TOKEN:
                 payload['keepAlive'] = {'type': 'empty'}
@@ -576,7 +669,7 @@ class Client:
                 'Authorization': "jwt %s" % self.auth.get("jwt")
             }
 
-        raise DevoClientException(_format_error(ERROR_MSGS['no_auth']))
+        raise DevoClientException((ERROR_MSGS['no_auth']))
 
     def _get_sign(self, data, tstamp):
         """
@@ -587,8 +680,7 @@ class Client:
         """
         if not self.auth.get("key", False) \
                 or not self.auth.get("secret", False):
-            raise DevoClientException(_format_error("You need a API Key and "
-                                                    "API secret to make this"))
+            raise DevoClientException(ERROR_MSGS["missing_api_key"])
         sign = hmac.new(self.auth.get("secret").encode("utf-8"),
                         (self.auth.get("key") + data + tstamp).encode("utf-8"),
                         hashlib.sha256)
@@ -678,5 +770,141 @@ class Client:
                 except json.decoder.JSONDecodeError:
                     return response.text
             tries += 1
-            time.sleep(self.timeout)
+            time.sleep(self.retry_delay * (2 ** (tries-1)))
         return {}
+
+    @staticmethod
+    def _future_queries_available(resp):
+        """
+        Verify whether future queries are supported for the response type
+        selected
+        :param resp: str
+        :return: bool
+        """
+        return resp not in ["msgpack", "xls"]
+
+    @staticmethod
+    def _fromDate_parser(fromDate, now=datetime.now()):
+
+        if isinstance(fromDate, (float, int)):
+            return fromDate
+        now = now.astimezone(pytz.UTC)
+        now_milliseconds = now.timestamp() * 1000
+        adate = datetime.strptime(str(now.date()), '%Y-%m-%d').replace(
+            tzinfo=pytz.utc)
+
+        if re.match('^[1-9]+(d|ad|h|ah|m|am|s|as)', fromDate):
+            date = re.split('(d|ad|h|ah|m|am|s|as)', fromDate)
+            num = int(date[0])
+            unit = date[1]
+
+            if unit == "d":
+                return now_milliseconds - (8.64e+7 * num)
+            elif unit == "ad":
+                return adate.timestamp() * 1000 - (8.64e+7 * num)
+            elif unit == "h":
+                return now_milliseconds - (3.6e+6 * num)
+            elif unit == "ah":
+                return adate.timestamp() * 1000 - (3.6e+6 * num)
+            elif unit == "m":
+                return now_milliseconds - (60000 * num)
+            elif unit == "am":
+                return adate.timestamp() * 1000 - (60000 * num)
+            elif unit == "s":
+                return now_milliseconds - (1000 * num)
+            elif unit == "as":
+                return adate.timestamp() * 1000 - (1000 * num)
+
+        elif fromDate == "today":
+            return adate.timestamp() * 1000
+        elif fromDate == "endday":
+            return (adate + timedelta(hours=23, minutes=59,
+                                      seconds=59)).timestamp() * 1000
+        elif fromDate == "endmonth":
+            return (adate.replace(
+                day=calendar.monthrange(adate.year, adate.month)[
+                    1]) + timedelta(hours=23, minutes=59,
+                                    seconds=59)).timestamp() * 1000
+        elif fromDate == "now":
+            return now.timestamp() * 1000
+
+    @staticmethod
+    def _toDate_parser(fromDateMillisec, toDate, now=datetime.now()):
+
+        if isinstance(toDate, (float, int)):
+            return toDate
+
+        now = now.astimezone(pytz.UTC)
+        fromDate = datetime.fromtimestamp(fromDateMillisec / 1000).replace(
+            tzinfo=pytz.utc)
+        aFromdate = datetime.strptime(str(fromDate.date()),
+                                      '%Y-%m-%d').replace(tzinfo=pytz.utc)
+        aNowdate = datetime.strptime(str(now.date()), '%Y-%m-%d').replace(
+            tzinfo=pytz.utc)
+
+        if re.match('^[1-9]+(d|ad|h|ah|m|am|s|as)', toDate):
+            date = re.split('(d|ad|h|ah|m|am|s|as)', toDate)
+            num = int(date[0])
+            unit = date[1]
+
+            if unit == "d":
+                return fromDateMillisec + (8.64e+7 * num)
+            elif unit == "ad":
+                return aFromdate.timestamp() * 1000 + (8.64e+7 * num)
+            elif unit == "h":
+                return fromDateMillisec + (3.6e+6 * num)
+            elif unit == "ah":
+                return aFromdate.timestamp() * 1000 + (3.6e+6 * num)
+            elif unit == "m":
+                return fromDateMillisec + (60000 * num)
+            elif unit == "am":
+                return aFromdate.timestamp() * 1000 + (60000 * num)
+            elif unit == "s":
+                return fromDateMillisec + (1000 * num)
+            elif unit == "as":
+                return aFromdate.timestamp() * 1000 + (1000 * num)
+
+        elif toDate == "today":
+            return aNowdate.timestamp() * 1000
+        elif toDate == "endday":
+            return (aFromdate + timedelta(hours=23, minutes=59,
+                                          seconds=59)).timestamp() * 1000
+        elif toDate == "endmonth":
+            return (aFromdate.replace(
+                day=calendar.monthrange(aFromdate.year, aFromdate.month)[
+                    1]) + timedelta(hours=23, minutes=59,
+                                    seconds=59)).timestamp() * 1000
+        elif toDate == "now":
+            return now.timestamp() * 1000
+
+    def _error_handler(self, content):
+        """
+        Function to manage possible errors returned from malote queries that
+        Serrea is sending as part of the http response
+        Depending on the response output format the error will be handled
+        by different ways
+        """
+        if self.config.response in ["xls", "msgpack"]:
+            return content
+        else:
+            pattern = DATA_ERROR_PATTERNS[self.config.response]
+            match = re.search(pattern, content)
+            if match:
+                error = match.group(0)
+                code = int(match.group(1))
+                message = match.group(2).strip()
+                raise DevoClientException(
+                    ERROR_MSGS["data_query_error"]
+                    % message, code=code, cause=error)
+            else:
+                return content
+
+
+DATA_ERROR_PATTERNS = {
+    "json/simple/compact": r'{"e":\[(\d+),"([^"]+)"]}',
+    "json/simple": r'\["error",(\d+),"([^"]+)"\]',
+    "json": r'"error":\s\[(\d+),"([^"]+)"\]',
+    "json/compact": r'"e":\s\[(\d+),"([^"]+)"\]',
+    "csv": r'devo\.api\.error,(\d+),([^,\n]+)',
+    'tsv': r'devo\.api\.error(?:\t|\s+)(\d+)(?:\t|\s+)([^\t\n]+)'
+}
